// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl.html       *
// * Copyright (C) 2008-2011 ZenJu (zhnmju123 AT gmx.de)                    *
// **************************************************************************

#include "file_op.h"
#include <algorithm>
#include <string>
#define WIN32_LEAN_AND_MEAN
#include <zen/com_ptr.h>
#include <zen/com_error.h>

#include <shellapi.h> //shell constants such as FO_* values
#include <shobjidl.h>

using namespace zen;


namespace
{
void moveToRecycleBin(const wchar_t* fileNames[],    //throw ComError
                      size_t         fileNo) //size of fileNames array
{
    ComPtr<IFileOperation> fileOp;
    ZEN_CHECK_COM(::CoCreateInstance(CLSID_FileOperation, //throw ComError
                                     NULL,
                                     CLSCTX_ALL,
                                     IID_PPV_ARGS(fileOp.init())));

    // Set the operation flags.  Turn off  all UI
    // from being shown to the user during the
    // operation.  This includes error, confirmation
    // and progress dialogs.
    ZEN_CHECK_COM(fileOp->SetOperationFlags(FOF_ALLOWUNDO | //throw ComError
                                            FOF_NOCONFIRMATION |
                                            FOF_SILENT         |
                                            FOFX_EARLYFAILURE  |
                                            FOF_NOERRORUI));

    int operationCount = 0;

    for (size_t i = 0; i < fileNo; ++i)
    {
        //create file/folder item object
        ComPtr<IShellItem> psiFile;
        HRESULT hr = ::SHCreateItemFromParsingName(fileNames[i],
                                                   NULL,
                                                   IID_PPV_ARGS(psiFile.init()));
        if (FAILED(hr))
        {
            if (hr == HRESULT_FROM_WIN32(ERROR_FILE_NOT_FOUND) || //file not existing anymore
                hr == HRESULT_FROM_WIN32(ERROR_PATH_NOT_FOUND))
                continue;
            throw ComError(std::wstring(L"Error calling \"SHCreateItemFromParsingName\" for file:\n") + L"\"" + fileNames[i] + L"\".", hr);
        }

        ZEN_CHECK_COM(fileOp->DeleteItem(psiFile.get(), NULL));

        ++operationCount;
    }

    if (operationCount == 0) //calling PerformOperations() without anything to do would result in E_UNEXPECTED
        return;

    //perform actual operations
    ZEN_CHECK_COM(fileOp->PerformOperations());

    //check if errors occured: if FOFX_EARLYFAILURE is not used, PerformOperations() can return with success despite errors!
    BOOL pfAnyOperationsAborted = FALSE;
    ZEN_CHECK_COM(fileOp->GetAnyOperationsAborted(&pfAnyOperationsAborted));

    if (pfAnyOperationsAborted == TRUE)
        throw ComError(L"Operation did not complete successfully.");
}


void copyFile(const wchar_t* sourceFile, //throw ComError
              const wchar_t* targetFile)
{
    ComPtr<IFileOperation> fileOp;
    ZEN_CHECK_COM(::CoCreateInstance(CLSID_FileOperation, //throw ComError
                                     NULL,
                                     CLSCTX_ALL,
                                     IID_PPV_ARGS(fileOp.init())));

    // Set the operation flags.  Turn off  all UI
    // from being shown to the user during the
    // operation.  This includes error, confirmation
    // and progress dialogs.
    ZEN_CHECK_COM(fileOp->SetOperationFlags(FOF_NOCONFIRMATION | //throw ComError
                                            FOF_SILENT         |
                                            FOFX_EARLYFAILURE  |
                                            FOF_NOERRORUI));
    //create source object
    ComPtr<IShellItem> psiSourceFile;
    {
        HRESULT hr = ::SHCreateItemFromParsingName(sourceFile,
                                                   NULL,
                                                   IID_PPV_ARGS(psiSourceFile.init()));
        if (FAILED(hr))
            throw ComError(std::wstring(L"Error calling \"SHCreateItemFromParsingName\" for file:\n") + L"\"" + sourceFile + L"\".", hr);
    }

    const size_t pos = std::wstring(targetFile).find_last_of(L'\\');
    if (pos == std::wstring::npos)
        throw ComError(L"Target filename does not contain a path separator.");

    const std::wstring targetFolder(targetFile, pos);
    const std::wstring targetFileNameShort = targetFile + pos + 1;

    //create target folder object
    ComPtr<IShellItem> psiTargetFolder;
    {
        HRESULT hr = ::SHCreateItemFromParsingName(targetFolder.c_str(),
                                                   NULL,
                                                   IID_PPV_ARGS(psiTargetFolder.init()));
        if (FAILED(hr))
            throw ComError(std::wstring(L"Error calling \"SHCreateItemFromParsingName\" for folder:\n") + L"\"" + targetFolder + L"\".", hr);
    }

    //schedule file copy operation
    ZEN_CHECK_COM(fileOp->CopyItem(psiSourceFile.get(), psiTargetFolder.get(), targetFileNameShort.c_str(), NULL));

    //perform actual operations
    ZEN_CHECK_COM(fileOp->PerformOperations());

    //check if errors occured: if FOFX_EARLYFAILURE is not used, PerformOperations() can return with success despite errors!
    BOOL pfAnyOperationsAborted = FALSE;
    ZEN_CHECK_COM(fileOp->GetAnyOperationsAborted(&pfAnyOperationsAborted));

    if (pfAnyOperationsAborted == TRUE)
        throw ComError(L"Operation did not complete successfully.");
}


inline
void copyString(const std::wstring& input, wchar_t* buffer, size_t bufferSize)
{
    if (bufferSize > 0)
    {
        //size_t endPos = input.copy(buffer, bufferSize - 1);
        //buffer[endPos] = 0;
        const size_t maxSize = std::min(input.length(), bufferSize - 1);
        std::copy(input.begin(), input.begin() + maxSize, buffer);
        buffer[maxSize] = 0;
    }
}

std::wstring lastErrorMessage; //this should really be thread-local!!!
}


bool fileop::moveToRecycleBin(const wchar_t* fileNames[],
                              size_t         fileNo) //size of fileNames array
{
    try
    {
        ::moveToRecycleBin(fileNames, fileNo); //throw ComError
        return true;
    }
    catch (const zen::ComError& e)
    {
        lastErrorMessage = e.toString();
        return false;
    }
}


bool fileop::copyFile(const wchar_t* sourceFile,
                      const wchar_t* targetFile)
{
    try
    {
        ::copyFile(sourceFile, targetFile); //throw ComError
        return true;
    }
    catch (const zen::ComError& e)
    {
        lastErrorMessage = e.toString();
        return false;
    }
}


//if any of the functions above returns 'false', this message returns last error
void fileop::getLastError(wchar_t* errorMessage, size_t errorBufferLen)
{
    copyString(lastErrorMessage, errorMessage, errorBufferLen);
}
